/* simann.f -- translated by f2c (version 20020621). */

/* ABSTRACT: */

/* Simulated annealing is a global optimization method that distinguishes    */
/* between different local optima. Starting from an initial point, the       */
/* algorithm takes a step and the function is evaluated. When minimizing a   */
/* function, any downhill step is accepted and the process repeats from this */
/* new point. An uphill step may be accepted. Thus, it can escape from local */
/* optima. This uphill decision is made by the Metropolis criteria. As the   */
/* optimization process proceeds, the length of the steps decline and the    */
/* algorithm closes in on the global optimum. Since the algorithm makes very */
/* few assumptions regarding the function to be optimized, it is quite       */
/* robust with respect to non-quadratic surfaces. The degree of robustness   */
/* can be adjusted by the user. In fact, simulated annealing can be used as  */
/* a local optimizer for difficult functions.                                */

/* This implementation of simulated annealing was used in                    */
/* "Global Optimization of Statistical Functions with Simulated Annealing,"  */
/* Goffe, Ferrier and Rogers, Journal of Econometrics, vol. 60, no. 1/2,     */
/* Jan./Feb. 1994, pp. 65-100. Briefly, we found it competitive, if not      */
/* superior, to multiple restarts of conventional optimization routines for  */
/* difficult optimization problems.                                          */

/* For more information on this routine, contact its author:                 */
/* Bill Goffe, bgoffe@whale.st.usm.edu                                       */

#include <stdio.h>
#include <math.h>

#define FALSE_ 0
#define TRUE_  1

/* Common Block Declarations */

struct {
  float u[97], c_, cd, cm;
  int i97, j97;
} raset;

/* Table of constant values */

static int c_2 = 2;
static int c_4 = 4;

/* Main program */ 

/* This file is an example of the Corana et al. simulated annealing         */
/* algorithm for multimodal and robust optimization as implemented          */
/* and modified by Goffe, Ferrier and Rogers. Counting the above line       */
/* ABSTRACT as 1, the routine itself (SA), with its supplementary           */
/* routines, is on lines 232-990. A multimodal example from Judge et al.    */
/* (FCN) is on lines 150-231. The rest of this file (lines 1-149) is a      */
/* driver routine with values appropriate for the Judge example. Thus, this */
/* example is ready to run.                                                 */

/* To understand the algorithm, the documentation for SA on lines 236-484   */
/* should be read along with the parts of the paper that describe           */
/* simulated annealing. Then the following lines will then aid the user     */
/* in becomming proficient with this implementation of simulated annealing. */

/* Learning to use SA:                                                      */

/* Use the sample function from Judge with the following suggestions        */
/* to get a feel for how SA works. When you've done this, you should be     */
/* ready to use it on most any function with a fair amount of expertise.    */

/* 1. Run the program as is to make sure it runs okay. Take a look at       */
/*    the intermediate output and see how it optimizes as temperature       */
/*    (T) falls. Notice how the optimal point is reached and how            */
/*    falling T reduces VM.                                                 */

/* 2. Look through the documentation to SA so the following makes a         */
/*    bit of sense. In line with the paper, it shouldn't be that hard       */
/*    to figure out. The core of the algorithm is described on pp. 68-70    */
/*    and on pp. 94-95. Also see Corana et al. pp. 264-9.                   */

/* 3. To see how it selects points and makes decisions about uphill         */
/*    and downhill moves, set IPRINT = 3 (very detailed intermediate        */
/*    output) and MAXEVL = 100 (only 100 function evaluations to limit      */
/*    output).                                                              */

/* 4. To see the importance of different temperatures, try starting         */
/*    with a very low one (say T = 10E-5). You'll see (i) it never          */
/*    escapes from the local optima (in annealing terminology, it           */
/*    quenches) & (ii) the step length (VM) will be quite small. This       */
/*    is a key part of the algorithm: as temperature (T) falls, step        */
/*    length does too. In a minor point here, note how VM is quickly        */
/*    reset from its initial value. Thus, the input VM is not very          */
/*    important. This is all the more reason to examine VM once the         */
/*    algorithm is underway.                                                */

/* 5. To see the effect of different parameters and their effect on         */
/*    the speed of the algorithm, try RT = .95 & RT = .1. Notice the        */
/*    vastly different speed for optimization. Also try NT = 20. Note       */
/*    that this sample function is quite easy to optimize, so it will       */
/*    tolerate big changes in these parameters. RT and NT are the           */
/*    parameters one should adjust to modify the runtime of the             */
/*    algorithm and its robustness.                                         */

/* 6. Try constraining the algorithm with either LB or UB.                  */

/*  Set underflows to zero on IBM mainframes.                               */
/*  CALL XUFLOW(0)                                                          */

// int main(void)

int xsimann()
{
  /* Local variables */
  
  static int i_;
  static int ns, nt;
  static int ier;
  static int max_;
  static int nacc, nacp[2];
  static int nobds;
  static int iseed1, iseed2, nfcnev, maxevl;
  static int iprint;
  static double c_[2];
  static double t, x[2], lb[2];
  static double ub[2];
  static double vm[2], rt, xp[2];
  static double eps;
  static double fopt, xopt[2];
  static double fstar[4];
  
  /* Subroutines */ 
  
  extern int sa(int *, double *, int *, double *, double *, int *, int *, 
		int *, int *, double *, double *, double *, int *, int *, 
		int *, double *, double *, double *, double *, int *, int *, 
		int *, int *, double *, double *, int *);
  extern int prtvec(double *, int *, char *);
  
  /*  Set input parameters. */
  
  max_ = FALSE_;
  eps = 1e-6;
  rt = 0.5f;
  iseed1 = 1;
  iseed2 = 2;
  ns = 20;
  nt = 5;
  maxevl = 100000;
  iprint = 1;
  
  for(i_ = 1; i_ <= 2; ++i_) {
    lb[i_ - 1] = - 1e25;
    ub[i_ - 1] = 1e25;
    c_[i_ - 1] = 2.0f;
  }

  /*  Note start at local, but not global, optima of the Judge function. */
 
  x[0] = 2.354471f;
  x[1] = -0.319186f;
  
  /*  Set input values of the input/output parameters. */
  
  t = 5.0f;
  
  for(i_ = 1; i_ <= 2; ++i_) {
    vm[i_ - 1] = 1.0f;
  }

  printf("SIMULATED ANNEALING EXAMPLE \n\n");
  printf("NUMBER OF PARAMETERS: %d \n", c_2);
  printf("MAXIMAZATION:         %d \n", max_);
  printf("INITIAL TEMP:         %lf \n", t);
  printf("RT:                   %lf \n", rt);
  printf("EPS:                  %lf \n", eps);
  printf("NS:                   %d \n", ns);
  printf("NT:                   %d \n", nt);
  printf("NEPS:                 %d \n", c_4);
  printf("MAXEVL:               %d \n", maxevl);
  printf("IPRINT:               %d \n", iprint);
  printf("ISEED1:               %d \n", iseed1);
  printf("ISEED2:               %d \n", iseed2);
  
  prtvec(x, &c_2, "STARTING VALUES");
  prtvec(vm, &c_2, "INITIAL STEP LENGTH");
  prtvec(lb, &c_2, "LOWER BOUND");
  prtvec(ub, &c_2, "UPPER BOUND");
  prtvec(c_, &c_2, "C VECTOR");

  printf(" ****   END OF DRIVER ROUTINE OUTPUT   **** \n");
  printf(" ****   BEFORE CALL TO SA.             **** \n");      

  sa(&c_2, x, &max_, &rt, &eps, &ns, &nt, &c_4, &maxevl, lb, ub, c_, &iprint, 
     &iseed1, &iseed2, &t, vm, xopt, &fopt, &nacc, &nfcnev, &nobds, &ier, 
     fstar, xp, nacp);
  
  printf(" ****   RESULTS AFTER SA   **** \n");  

  prtvec(xopt, &c_2, "SOLUTION");
  prtvec(vm, &c_2, "FINAL STEP LENGTH");
  
  printf("OPTIMAL FUNCTION VALUE:             %lf \n", fopt);
  printf("NUMBER OF FUNCTION EVALUATIONS:     %d \n", nfcnev);
  printf("NUMBER OF ACCEPTED EVALUATIONS:     %d \n", nacc);
  printf("NUMBER OF OUT OF BOUND EVALUATIONS: %d \n", nobds);
  printf("FINAL TEMP:                         %lf \n", t);
  printf("IER:                                %d \n", ier);
  
  return(0);
} /* MAIN */

/* Subroutine */ 

/*  This subroutine is from the example in Judge et al., The Theory and      */
/*  Practice of Econometrics, 2nd ed., pp. 956-7. There are two optima:      */
/*  F(0.864, 1.23) = 16.0817 (the global minumum) and                        */
/*  F(2.35, -0.319) = 20.9805.                                               */

int fcn(int *n, double *theta, double *h_)
{
  /* System generated locals */

  double d_1, d_2;
  
  /* Local variables */
  
  static int i_;
  static double y[20], x2[20], x3[20];
  
  /* Parameter adjustments */
  
  --theta;
  
  /* Function Body */
  
  y[0]   = 4.284f;
  y[1]   = 4.149f;
  y[2]   = 3.877f;
  y[3]   = 0.533f;
  y[4]   = 2.211f;
  y[5]   = 2.389f;
  y[6]   = 2.145f;
  y[7]   = 3.231f;
  y[8]   = 1.998f;
  y[9]   = 1.379f;
  y[10]  = 2.106f;
  y[11]  = 1.428f;
  y[12]  = 1.011f;
  y[13]  = 2.179f;
  y[14]  = 2.858f;
  y[15]  = 1.388f;
  y[16]  = 1.651f;
  y[17]  = 1.593f;
  y[18]  = 1.046f;
  y[19]  = 2.152f;
  x2[0]  = 0.286f;
  x2[1]  = 0.973f;
  x2[2]  = 0.384f;
  x2[3]  = 0.276f;
  x2[4]  = 0.973f;
  x2[5]  = 0.543f;
  x2[6]  = 0.957f;
  x2[7]  = 0.948f;
  x2[8]  = 0.543f;
  x2[9]  = 0.797f;
  x2[10] = 0.936f;
  x2[11] = 0.889f;
  x2[12] = 0.006f;
  x2[13] = 0.828f;
  x2[14] = 0.399f;
  x2[15] = 0.617f;
  x2[16] = 0.939f;
  x2[17] = 0.784f;
  x2[18] = 0.072f;
  x2[19] = 0.889f;
  x3[0]  = 0.645f;
  x3[1]  = 0.585f;
  x3[2]  = 0.31f;
  x3[3]  = 0.058f;
  x3[4]  = 0.455f;
  x3[5]  = 0.779f;
  x3[6]  = 0.259f;
  x3[7]  = 0.202f;
  x3[8]  = 0.028f;
  x3[9]  = 0.099f;
  x3[10] = 0.142f;
  x3[11] = 0.296f;
  x3[12] = 0.175f;
  x3[13] = 0.18f;
  x3[14] = 0.842f;
  x3[15] = 0.039f;
  x3[16] = 0.103f;
  x3[17] = 0.62f;
  x3[18] = 0.158f;
  x3[19] = 0.704f;
  *h_ = 0.0f;
  
  for(i_ = 1; i_ <= 20; ++i_) {
    /* Computing 2nd power */
	
    d_2 = theta[2];

    /* Computing 2nd power */
    
    d_1 = theta[1] + theta[2] * x2[i_ - 1] + d_2 * d_2 * x3[i_ - 1] 
      - y[i_ - 1];
    *h_ = d_1 * d_1 + *h_;
  }
  
  return(0);
} /* fcn */

/* Subroutine */ 

/* Version: 3.2                                                           */

/* Date: 1/22/94.                                                          */

/* Differences compared to Version 2.0:                                    */

/* 1. If a trial is out of bounds, a point is randomly selected            */
/*    from LB(i) to UB(i). Unlike in version 2.0, this trial is            */
/*    evaluated and is counted in acceptances and rejections.              */
/*    All corresponding documentation was changed as well.                 */ 

/* Differences compared to Version 3.0:                                    */

/* 1. If VM(i) > (UB(i) - LB(i)), VM is set to UB(i) - LB(i).              */
/*    The idea is that if T is high relative to LB & UB, most              */
/*    points will be accepted, causing VM to rise. But, in this            */
/*    situation, VM has little meaning; particularly if VM is              */
/*    larger than the acceptable region. Setting VM to this size           */
/*    still allows all parts of the allowable region to be selected.       */

/* Differences compared to Version 3.1:                                    */

/*  1. Test made to see if the initial temperature is positive.            */
/*  2. WRITE statements prettied up.                                       */
/*  3. References to paper updated.                                        */

/* Synopsis:                                                               */

/* This routine implements the continuous simulated annealing global       */
/* optimization algorithm described in Corana et al.'s article             */
/* "Minimizing Multimodal Functions of Continuous Variables with the       */
/* "Simulated Annealing" Algorithm" in the September 1987 (vol. 13, # 3,   */
/* pp. 262-280) issue of the ACM Transactions on Mathematical Software.    */

/* A very quick (perhaps too quick) overview of SA:                        */

/* SA tries to find the global optimum of an N dimensional function.       */

/* It moves both up and downhill and as the optimization process           */
/* proceeds, it focuses on the most promising area.                        */

/* To start, it randomly chooses a trial point within the step length      */
/* VM (a vector of length N) of the user selected starting point. The      */
/* function is evaluated at this trial point and its value is compared     */
/* to its value at the initial point.                                      */

/* In a maximization problem, all uphill moves are accepted and the        */
/* algorithm continues from that trial point. Downhill moves may be        */
/* accepted; the decision is made by the Metropolis criteria. It uses T    */
/* (temperature) and the size of the downhill move in a probabilistic      */
/* manner. The smaller T and the size of the downhill move are, the more   */
/* likely that move will be accepted. If the trial is accepted, the        */
/* algorithm moves on from that point. If it is rejected, another point    */
/* is chosen instead for a trial evaluation.                               */

/* Each element of VM periodically adjusted so that half of all            */
/* function evaluations in that direction are accepted.                    */

/* A fall in T is imposed upon the system with the RT variable by          */
/* T(i+1) = RT*T(i) where i is the ith iteration. Thus, as T declines,     */
/* downhill moves are less likely to be accepted and the percentage of     */
/* rejections rise. Given the scheme for the selection for VM, VM falls.   */
/* Thus, as T declines, VM falls and SA focuses upon the most promising    */
/* area for optimization.                                                  */

/* The importance of the parameter T:                                      */

/* The parameter T is crucial in using SA successfully. It influences      */
/* VM, the step length over which the algorithm searches for optima. For   */
/* a small intial T, the step length may be too small; thus not enough     */
/* of the function might be evaluated to find the global optima. The user  */
/* should carefully examine VM in the intermediate output (set IPRINT = 1) */ 
/* to make sure that VM is appropriate. The relationship between the       */
/* initial temperature and the resulting step length is function           */
/* dependent.                                                              */

/* To determine the starting temperature that is consistent with           */
/* optimizing a function, it is worthwhile to run a trial run first. Set   */
/* RT = 1.5 and T = 1.0. With RT > 1.0, the temperature increases and VM   */
/* rises as well. Then select the T that produces a large enough VM.       */

/* For modifications to the algorithm and many details on its use,         */
/* (particularly for econometric applications) see Goffe, Ferrier          */
/* and Rogers, "Global Optimization of Statistical Functions with          */
/* Simulated Annealing," Journal of Econometrics, vol. 60, no. 1/2,        */
/* Jan./Feb. 1994, pp. 65-100.                                             */

/* For more information, contact                                           */

/*          Bill Goffe                                                     */
/*          Department of Economics and International Business             */
/*          University of Southern Mississippi                             */
/*          Hattiesburg, MS  39506-5072                                    */
/*          (601) 266-4484 (office)                                        */
/*          (601) 266-4920 (fax)                                           */
/*          bgoffe@whale.st.usm.edu (Internet)                             */

/* As far as possible, the parameters here have the same name as in        */
/* the description of the algorithm on pp. 266-8 of Corana et al.          */

/* In this description, SP is single precision, DP is double precision,    */
/* INT is int, L is logical and (N) denotes an array of length n.          */
/* Thus, DP(N) denotes a double precision array of length n.               */

/* Input Parameters:                                                       */

/* Note: The suggested values generally come from Corana et al. To         */
/*       drastically reduce runtime, see Goffe et al., pp. 90-1 for        */
/*       suggestions on choosing the appropriate RT and NT.                */

/* N      - Number of variables in the function to be optimized. (INT)     */
/* X      - The starting values for the variables of the function to be    */
/*          optimized. (DP(N))                                             */
/* MAX    - Denotes whether the function should be maximized or            */
/*          minimized. A true value denotes maximization while a false     */
/*          value denotes minimization. Intermediate output (see IPRINT)   */
/*          takes this into account. (L)                                   */
/* RT     - The temperature reduction factor. The value suggested by       */
/*          Corana et al. is .85. See Goffe et al. for more advice. (DP)   */
/* EPS    - Error tolerance for termination. If the final function         */
/*          values from the last neps temperatures differ from the         */
/*          corresponding value at the current temperature by less than    */
/*          EPS and the final function value at the current temperature    */
/*          differs from the current optimal function value by less than   */
/*          EPS, execution terminates and IER = 0 is returned. (EP)        */
/* NS     - Number of cycles. After NS*N function evaluations, each        */
/*          element of VM is adjusted so that approximately half of        */
/*          all function evaluations are accepted. The suggested value     */
/*          is 20. (INT)                                                   */
/* NT     - Number of iterations before temperature reduction. After       */
/*          NT*NS*N function evaluations, temperature (T) is changed       */
/*          by the factor RT. Value suggested by Corana et al. is          */
/*          MAX(100, 5*N). See Goffe et al. for further advice. (INT)      */
/* NEPS   - Number of final function values used to decide upon termi-     */
/*          nation. See EPS. Suggested value is 4. (INT)                   */
/* MAXEVL - The maximum number of function evaluations. If it is           */
/*          exceeded, IER = 1. (INT)                                       */
/* LB     - The lower bound for the allowable solution variables. (DP(N))  */
/* UB     - The upper bound for the allowable solution variables. (DP(N))  */
/*          If the algorithm chooses X(I) .LT. LB(I) or X(I) .GT. UB(I),   */
/*          I = 1, N, a point is from inside is randomly selected. This    */
/*          This focuses the algorithm on the region inside UB and LB.     */
/*          Unless the user wishes to concentrate the search to a par-     */
/*          ticular region, UB and LB should be set to very large positive */
/*          and negative values, respectively. Note that the starting      */
/*          vector X should be inside this region. Also note that LB and   */
/*          UB are fixed in position, while VM is centered on the last     */
/*          accepted trial set of variables that optimizes the function.   */
/* C      - Vector that controls the step length adjustment. The suggested */
/*          value for all elements is 2.0. (DP(N))                         */
/* IPRINT - controls printing inside SA. (INT)                             */
/*          Values: 0 - Nothing printed.                                   */
/*                  1 - Function value for the starting value and          */
/*                      summary results before each temperature            */
/*                      reduction. This includes the optimal               */
/*                      function value found so far, the total             */
/*                      number of moves (broken up into uphill,            */
/*                      downhill, accepted and rejected), the              */
/*                      number of out of bounds trials, the                */
/*                      number of new optima found at this                 */
/*                      temperature, the current optimal X and             */
/*                      the step length VM. Note that there are            */
/*                      N*NS*NT function evalutations before each          */
/*                      temperature reduction. Finally, notice is          */
/*                      is also given upon achieveing the termination      */
/*                      criteria.                                          */
/*                  2 - Each new step length (VM), the current optimal     */
/*                      X (XOPT) and the current trial X (X). This         */
/*                      gives the user some idea about how far X           */
/*                      strays from XOPT as well as how VM is adapting     */
/*                      to the function.                                   */
/*                  3 - Each function evaluation, its acceptance or        */
/*                      rejection and new optima. For many problems,       */
/*                      this option will likely require a small tree       */
/*                      if hard copy is used. This option is best          */
/*                      used to learn about the algorithm. A small         */
/*                      value for MAXEVL is thus recommended when          */
/*                      using IPRINT = 3.                                  */
/*                  Suggested value: 1                                     */
/*                  Note: For a given value of IPRINT, the lower valued    */
/*                        options (other than 0) are utilized.             */
/* ISEED1 - The first seed for the random number generator RANMAR.         */
/*          0 .LE. ISEED1 .LE. 31328. (INT)                                */
/* ISEED2 - The second seed for the random number generator RANMAR.        */
/*          0 .LE. ISEED2 .LE. 30081. Different values for ISEED1          */
/*          and ISEED2 will lead to an entirely different sequence         */
/*          of trial points and decisions on downhill moves (when          */
/*          maximizing). See Goffe et al. on how this can be used          */
/*          to test the results of SA. (INT)                               */

/* Input/Output Parameters:                                                */

/* T  - On input, the initial temperature. See Goffe et al. for advice.    */
/*      On output, the final temperature. (DP)                             */
/* VM - The step length vector. On input it should encompass the           */
/*      region of interest given the starting value X. For point           */
/*      X(I), the next trial point is selected is from X(I) - VM(I)        */
/*      to  X(I) + VM(I). Since VM is adjusted so that about half          */
/*      of all points are accepted, the input value is not very            */
/*      important (i.e. is the value is off, SA adjusts VM to the          */
/*      correct value). (DP(N))                                            */

/* Output Parameters:                                                      */

/* XOPT   - The variables that optimize the function. (DP(N))              */
/* FOPT   - The optimal value of the function. (DP)                        */
/* NACC   - The number of accepted function evaluations. (INT)             */
/* NFCNEV - The total number of function evaluations. In a minor           */
/*          point, note that the first evaluation is not used in the       */
/*          core of the algorithm; it simply initializes the               */
/*          algorithm. (INT).                                              */
/* NOBDS  - The total number of trial function evaluations that            */
/*          would have been out of bounds of LB and UB. Note that          */
/*          a trial point is randomly selected between LB and UB.          */
/*          (INT)                                                          */
/* IER    - The error return number. (INT)                                 */
/*          Values: 0  - Normal return; termination criteria achieved.     */
/*                  1  - Number of function evaluations (NFCNEV) is        */
/*                       greater than the maximum number (MAXEVL).         */
/*                  2  - The starting value (X) is not inside the          */
/*                       bounds (LB and UB).                               */
/*                  3  - The initial temperature is not positive.          */
/*                  99 - Should not be seen; only used internally.         */

/* Work arrays that must be dimensioned in the calling routine:            */

/* RWK1 (DP(NEPS))  (FSTAR in SA)                                          */
/* RWK2 (DP(N))     (XP    "  " )                                          */
/* IWK  (INT(N))    (NACP  "  " )                                          */

/* Required Functions (included):                                          */

/* EXPREP - Replaces the function EXP to avoid under- and overflows.       */
/*          It may have to be modified for non IBM-type main-              */
/*          frames. (DP)                                                   */
/* RMARIN - Initializes the random number generator RANMAR.                */
/* RANMAR - The actual random number generator. Note that                  */
/*          RMARIN must run first (SA does this). It produces uniform      */
/*          random numbers on [0,1]. These routines are from               */
/*          Usenet's comp.lang.fortran. For a reference, see               */
/*          "Toward a Universal Random Number Generator"                   */
/*          by George Marsaglia and Arif Zaman, Florida State              */
/*          University Report: FSU-SCRI-87-50 (1987).                      */
/*          It was later modified by F. James and published in             */
/*          "A Review of Pseudo-random Number Generators." For             */
/*          further information, contact stuart@ads.com. These             */
/*          routines are designed to be portable on any machine            */
/*          with a 24-bit or more mantissa. I have found it produces       */
/*          identical results on a IBM 3081 and a Cray Y-MP.               */

/* Required Subroutines (included):                                        */

/* PRTVEC - Prints vectors.                                                */
/* PRT1 ... PRT10 - Prints intermediate output.                            */
/* FCN    - Function to be optimized. The form is                          */
/*          SUBROUTINE FCN(N, X, F)                                        */
/*          INT N                                                          */
/*          DOUBLE PRECISION  X(N), F                                      */
/*          ...                                                            */
/*          function code with F = F(X)                                    */
/*          ...                                                            */
/*          RETURN                                                         */
/*          END                                                            */
/*          Note: This is the same form used in the multivariable          */
/*          minimization algorithms in the IMSL edition 10 library.        */

/* Machine Specific Features:                                              */

/* 1. EXPREP may have to be modified if used on non-IBM type mainframes.   */
/*    Watch for under- and overflows in EXPREP.                            */
/* 2. Some FORMAT statements use G25.18; this may be excessive for         */
/*    some machines.                                                       */
/* 3. RMARIN and RANMAR are designed to be protable; they should not       */
/*    cause any problems.                                                  */
/*  Type all external variables.                                           */
/*  Type all internal variables.                                           */
/*  Type all functions.                                                    */
/*  Initialize the random number generator RANMAR.                         */

int sa(int *n, double *x, int *max_, double *rt, double *eps, int *ns, 
       int *nt, int *neps, int *maxevl, double *lb, double *ub, double *c_, 
       int *iprint, int *iseed1, int *iseed2, double *t, double *vm, 
       double *xopt, double *fopt, int *nacc, int *nfcnev, int *nobds, 
       int *ier, double *fstar, double *xp, int *nacp)
{
  /* System generated locals */
  
  int i_1, i_2, i_3, i_4;
  double d_1;
  
  /* Local variables */

  static int h_, i_, j, m;
  static int nup;  
  static int nrej;
  static int nnew;
  static int quit;
  static int ndown;
  static int lnobds;
  static double f;
  static double p, fp, pp;
  static double ratio;
 
  /* Subroutines */ 

  extern int fcn(int *, double *, double *);
  
  extern int prt1(void); 
  extern int prt2(int *, int *, double *, double *);
  extern int prt3(int *, int *, double *, double *, double *, double *);
  extern int prt4(int *, int *, double *, double *, double *, double *); 
  extern int prt5(void);
  extern int prt6(int *); 
  extern int prt7(int *); 
  extern int prt8(int *, double *, double *, double *);
  extern int prt9(int *, int *, double *, double *, double *, double *, 
		  int *, int *, int *, int *, int *);
  extern int prt10(void);
  extern double ranmar(void);
  extern int rmarin(int *, int *);
  extern double exprep(double *);
  
  /* Parameter adjustments */

  --nacp;
  --xp;
  --fstar;
  --xopt;
  --vm;
  --c_;
  --ub;
  --lb;
  --x;
  
  /* Function Body */
  
  rmarin(iseed1, iseed2);
  
  /*  Set initial values. */
  
  *nacc = 0;
  *nobds = 0;
  *nfcnev = 0;
  *ier = 99;
  i_1 = *n;
  
  for(i_ = 1; i_ <= i_1; ++i_) {
    xopt[i_] = x[i_];
    nacp[i_] = 0;
  }
  
  i_1 = *neps;
  
  for(i_ = 1; i_ <= i_1; ++i_) {
    fstar[i_] = 1e20;
  }
  
  /*  If the initial temperature is not positive, notify the user and */
  /*  return to the calling routine.                                  */
    
  if(*t <= 0.0f) {
    printf("THE INITIAL TEMPERATURE IS NOT POSITIVE. \n");
    printf("RESET THE VARIABLE T. \n");

    *ier = 3;
  
    return(0);
  }

  /*  If the initial value is out of bounds, notify the user and return */
  /*  to the calling routine. */

  i_1 = *n;
  
  for(i_ = 1; i_ <= i_1; ++i_) {
    if(x[i_] > ub[i_] || x[i_] < lb[i_]) {
      prt1();

      *ier = 2;
  
      return(0);
    }
  }

  /*  Evaluate the function with input X and return value as F. */
    
  fcn(n, &x[1], &f);
  
  /*  If the function is to be minimized, switch the sign of the function. */
  /*  Note that all intermediate and final output switches the sign back */
  /*  to eliminate any possible confusion for the user. */
  
  if(! (*max_)) {
    f = -f;
  }
  
  ++(*nfcnev);
  *fopt = f;
  fstar[1] = f;
  
  if(*iprint >= 1) {
    prt2(max_, n, &x[1], &f);
  }

  /*  Start the main loop. Note that it terminates if (i) the algorithm */
  /*  succesfully optimizes the function or (ii) there are too many */
  /*  function evaluations (more than MAXEVL). */

 L100:
  
  nup = 0;
  nrej = 0;
  nnew = 0;
  ndown = 0;
  lnobds = 0;
  i_1 = *nt;
  
  for(m = 1; m <= i_1; ++m) {
    i_2 = *ns;
    
    for(j = 1; j <= i_2; ++j) {
      i_3 = *n;
      
      for(h_ = 1; h_ <= i_3; ++h_) {
	/*  Generate XP, the trial value of X. Note use of VM to choose XP. */
	
	i_4 = *n;
	
	for(i_ = 1; i_ <= i_4; ++i_) {
	  if(i_ == h_) {
	    xp[i_] = x[i_] + (ranmar() * 2.f - 1.f) * vm[i_];
	  } else {
	    xp[i_] = x[i_];
	  }
	  
	  /* If XP is out of bounds, select a point in bounds for the trial. */
	  
	  if(xp[i_] < lb[i_] || xp[i_] > ub[i_]) {
	    xp[i_] = lb[i_] + (ub[i_] - lb[i_]) * ranmar();
	    ++lnobds;
	    ++(*nobds);
	    
	    if(*iprint >= 3) {
	      prt3(max_, n, &xp[1], &x[1], &fp, &f);
	    }
	  }
	}
	
	/*  Evaluate the function with the trial point XP and return as FP. */
	
	fcn(n, &xp[1], &fp);
	
	if(! (*max_)) {
	  fp = -fp;
	}
	
	++(*nfcnev);
	
	if(*iprint >= 3) {
	  prt4(max_, n, &xp[1], &x[1], &fp, &f);
	}
	
	/*  If too many function evaluations occur, terminate the algorithm. */
		
	if(*nfcnev >= *maxevl) {
	  prt5();
	  
	  if(!(*max_)) {
	    *fopt = -(*fopt);
	  }
	  
	  *ier = 1;
	  
	  return(0);
	}
	
	/*  Accept the new point if the function value increases. */
	
	if(fp >= f) {
	  if(*iprint >= 3) {
	    printf("POINT ACCEPTED \n");
	  }
	  
	  i_4 = *n;
		
	  for(i_ = 1; i_ <= i_4; ++i_) {
	    x[i_] = xp[i_];
	  }
		
	  f = fp;
	  ++(*nacc);
	  ++nacp[h_];
	  ++nup;
	  
	  /*  If greater than any other point, record as new optimum. */
	  
	  if(fp > *fopt) {
	    if(*iprint >= 3) {
	      printf("NEW OPTIMUM \n");
	    }
	    
	    i_4 = *n;
		
	    for(i_ = 1; i_ <= i_4; ++i_) {
	      xopt[i_] = xp[i_];
	    }
	    
	    *fopt = fp;
	    ++nnew;
	  }

	  /* If the point is lower, use the Metropolis criteria to decide on */
	  /*  acceptance or rejection. */
	} else {
	  d_1 = (fp - f) / *t;
	  p = exprep(&d_1);
	  pp = ranmar();
	  
	  if(pp < p) {
	    if(*iprint >= 3) {
	      prt6(max_);
	    }
	    
	    i_4 = *n;
		
	    for(i_ = 1; i_ <= i_4; ++i_) {
	      x[i_] = xp[i_];
	    }
		
	    f = fp;
	    ++(*nacc);
	    ++nacp[h_];
	    ++ndown;
	  } else {
	    ++nrej;
		
	    if (*iprint >= 3) {
	      prt7(max_);
	    }
	  }
	}
      }
    }
    
    /* Adjust VM so that approximately half of all evaluations are accepted. */
	
    i_2 = *n;
	
    for(i_ = 1; i_ <= i_2; ++i_) {
      ratio = (double) nacp[i_] / (double) (*ns);
	
      if(ratio > 0.6f) {
	vm[i_] *= c_[i_] * (ratio - 0.6f) / 0.4f + 1.0f;
      } else if (ratio < 0.4f) {
	vm[i_] /= c_[i_] * ((0.4f - ratio) / 0.4f) + 1.0f;
      }
      
      if(vm[i_] > ub[i_] - lb[i_]) {
	vm[i_] = ub[i_] - lb[i_];
      }
    }
    
    if(*iprint >= 2) {
      prt8(n, &vm[1], &xopt[1], &x[1]);
    }
	
    i_2 = *n;
    
    for(i_ = 1; i_ <= i_2; ++i_) {
      nacp[i_] = 0;
    }
  }
  
  if(*iprint >= 1) {
    prt9(max_, n, t, &xopt[1], &vm[1], fopt, &nup, &ndown, &nrej, &lnobds, 
	 &nnew);
  }

  /*  Check termination criteria. */
  
  quit = FALSE_;
  fstar[1] = f;
  
  if(*fopt - fstar[1] <= *eps) {
    quit = TRUE_;
  }
  
  i_1 = *neps;
  
  for(i_ = 1; i_ <= i_1; ++i_) {
    if((d_1 = f - fstar[i_], fabs(d_1)) > *eps) {
      quit = FALSE_;
    }
  }

  /*  Terminate SA if appropriate. */
  
  if(quit) {
    i_1 = *n;
    
    for(i_ = 1; i_ <= i_1; ++i_) {
      x[i_] = xopt[i_];
    }
	
    *ier = 0;
	
    if(!(*max_)) {
      *fopt = -(*fopt);
    }
	
    if(*iprint >= 1) {
	    prt10();
    }

    return(0);
  }
  
  /*  If termination criteria is not met, prepare for another loop. */
  
  *t = *rt * *t;
  
  for(i_ = *neps; i_ >= 2; --i_) {
    fstar[i_] = fstar[i_ - 1];
  }
  
  f = *fopt;
  i_1 = *n;
  
  for(i_ = 1; i_ <= i_1; ++i_) {
    x[i_] = xopt[i_];
  }
  
  /*  Loop again. */
  
  goto L100;
} /* sa */

/*  This function replaces exp to avoid under- and overflows and is    */
/*  designed for IBM 370 type machines. It may be necessary to modify  */
/*  it for other machines. Note that the maximum and minimum values of */
/*  EXPREP are such that they has no effect on the algorithm.          */

double exprep(double *rdum)
{
  /* System generated locals */
  
  double ret_val;
  
  /* Builtin functions */
  
  double exp(double);
    
  if(*rdum > 174.0f) {
    ret_val = 3.69e75;
  } else if (*rdum < -180.0f) {
    ret_val = 0.f;
  } else {
    ret_val = exp(*rdum);
  }
  
  return(ret_val);
} /* exprep */

/* Subroutine */ 

/* This subroutine and the next function generate random numbers. See      */
/* the comments for SA for more information. The only changes from the     */
/* orginal code is that (1) the test to make sure that RMARIN runs first   */
/* was taken out since SA assures that this is done (this test didn't      */
/* compile under IBM's VS Fortran) and (2) typing ivec as int was          */
/* taken out since ivec isn't used. With these exceptions, all following   */
/* lines are original.                                                     */

/* This is the initialization routine for the random number generator      */
/* RANMAR()                                                                */
/* NOTE: The seed variables can have values between:    0 <= IJ <= 31328   */
/*                                                      0 <= KL <= 30081   */

int rmarin(int *ij, int *kl)
{
  /* Local variables */
  
  static int i_, j, k, l, m;
  static float s, t;
  static int ii, jj;
  
  /* Fortran I/O blocks */
  
  if(*ij < 0 || *ij > 31328 || *kl < 0 || *kl > 30081) {
    printf("(A) The first random number seed must have a value"
	   " between 0 and 31328 \n");
    printf("(B) The second seed must have a value between 0 and 30081 \n");
  }
  
  i_ = *ij / 177 % 177 + 2;
  j = *ij % 177 + 2;
  k = *kl / 169 % 178 + 1;
  l = *kl % 169;
  
  for(ii = 1; ii <= 97; ++ii) {
    s = 0.f;
    t = .5f;
    
    for(jj = 1; jj <= 24; ++jj) {
      m = i_ * j % 179 * k % 179;
      i_ = j;
      j = k;
      k = m;
      l = (l * 53 + 1) % 169;
      
      if(l * m % 64 >= 32) {
	s += t;
      }
      
      t *= 0.5f;
    }
    
    raset.u[ii - 1] = s;
  }
  
  raset.c_  = 0.021602869033813477f;
  raset.cd  = 0.45623308420181274f;
  raset.cm  = 0.99999982118606567f;
  raset.i97 = 97;
  raset.j97 = 33;
  
  return(0);
} /* rmarin */

double ranmar(void)
{
  /* System generated locals */
  
  float ret_val;
  
  /* Local variables */
  
  static float uni;
  
  uni = raset.u[raset.i97 - 1] - raset.u[raset.j97 - 1];
  
  if(uni < 0.0f) {
    uni += 1.0f;
  }
  
  raset.u[raset.i97 - 1] = uni;
  --raset.i97;
  
  if(raset.i97 == 0) {
    raset.i97 = 97;
  }
  
  --raset.j97;
  
  if(raset.j97 == 0) {
    raset.j97 = 97;
  }
  
  raset.c_ -= raset.cd;
  
  if(raset.c_ < 0.0f) {
    raset.c_ += raset.cm;
  }
   
  uni -= raset.c_;
  
  if(uni < 0.0f) {
    uni += 1.0f;
  }
  
  ret_val = uni;
  
  return(ret_val);
} /* ranmar */

/* Subroutine */ 

/*  This subroutine prints intermediate output, as does PRT2 through   */
/*  PRT10. Note that if SA is minimizing the function, the sign of the */
/*  function value and the directions (up/down) are reversed in all    */
/*  output to correspond with the actual function optimization. This   */
/*  correction is because SA was written to maximize functions and     */
/*  it minimizes by maximizing the negative a function.                */

int prt1(void)
{
  printf("THE STARTING VALUE (X) IS OUTSIDE THE BOUNDS (LB AND UB). \n");
  printf("EXECUTION TERMINATED WITHOUT ANY \n");
  printf("OPTIMIZATION. RESPECIFY X, UB OR LB SO THAT \n");
  printf("LB(I) .LT. X(I) .LT. UB(I), I = 1, N. \n");  
  return(0);
} /* prt1 */

/* Subroutine */ 

int prt2(int *max_, int *n, double *x, double *f)
{
  /* System generated locals */

  double d_1;
  
  /* Subroutine */ 
  
  extern int prtvec(double *, int *, char *);
 
  /* Parameter adjustments */

  --x;
  
  /* Function Body */
  
  printf(" *** \n");
  prtvec(&x[1], n, "INITIAL X");
  
  if(*max_) {
    printf("INITIAL F: %lf \n", *f);
  } else {
    d_1 = -(*f);
    
    printf("INITIAL F: %lf \n", d_1);
  }
  
  return(0);
} /* prt2 */

/* Subroutine */ 

int prt3(int *max_, int *n, double *xp, double *x, double *fp, double *f)
{
  /* System generated locals */

  double d_1;
  
  /* Subroutine */ 

  extern int prtvec(double *, int *, char *);

  /* Parameter adjustments */
  
  --x;
  --xp;
  
  /* Function Body */

  printf(" *** \n");
  prtvec(&x[1], n, "CURRENT X");
  
  if(*max_) {
    printf("CURRENT F: %lf \n", *f);
  } else {
    d_1 = -(*f);
    
    printf("CURRENT F: %lf \n", d_1);
  }
  
  prtvec(&xp[1], n, "TRIAL X");
  printf("POINT REJECTED SINCE OUT OF BOUNDS \n");
  return(0);
} /* prt3 */

/* Subroutine */ 

int prt4(int *max_, int *n, double *xp, double *x, double *fp, double *f)
{
  /* System generated locals */
  
  double d_1;
  
  /* Subroutine */ 
  
  extern int prtvec(double *, int *, char *);
  
  /* Parameter adjustments */
  
  --x;
  --xp;
  
  /* Function Body */

  printf(" *** \n");
  prtvec(&x[1], n, "CURRENT X");
  
  if(*max_) {
    printf("CURRENT F:   %lf \n", *f); 
    prtvec(&xp[1], n, "TRIAL X");
    printf("RESULTING F: %lf \n", *fp);
  } else {
    d_1 = -(*f);
    
    printf("CURRENT F:   %lf \n", d_1);
    prtvec(&xp[1], n, "TRIAL X");
    
    d_1 = -(*fp);
    
    printf("RESULTING F: %lf \n", d_1);
  }

  return(0);
} /* prt4 */

/* Subroutine */ 

int prt5(void)
{
  printf("TOO MANY FUNCTION EVALUATIONS; CONSIDER INCREASING MAXEVL \n");
  printf("OR EPS, OR DECREASING NT OR RT. \n");
  printf("THESE RESULTS ARE LIKELY TO BE POOR. \n");
  return(0);
} /* prt5 */

/* Subroutine */ 

int prt6(int *max_)
{ 
  if(*max_) {
    printf("THOUGH LOWER, POINT ACCEPTED \n");
  } else {
    printf("THOUGH HIGHER, POINT ACCEPTED \n");
  }
  
  return(0);
} /* prt6 */

/* Subroutine */ 

int prt7(int *max_)
{
  if(*max_) {
    printf("LOWER POINT REJECTED \n");
  } else {
    printf("HIGHER POINT REJECTED \n");
  }
  
  return(0);
} /* prt7 */

/* Subroutine */ 

int prt8(int *n, double *vm, double *xopt, double *x)
{
  /* Subroutine */ 
  
  extern int prtvec(double *, int *, char *);
  
  /* Parameter adjustments */
  
  --x;
  --xopt;
  --vm;
  
  /* Function Body */

  printf("INTERMEDIATE RESULTS AFTER STEP LENGTH ADJUSTMENT \n");
  prtvec(&vm[1], n, "NEW STEP LENGTH (VM)");
  prtvec(&xopt[1], n, "CURRENT OPTIMAL X");
  prtvec(&x[1], n, "CURRENT X");
  printf(" *** \n");
  return(0);
} /* prt8 */

/* Subroutine */ 

int prt9(int *max_, int *n, double *t, double *xopt, double *vm, 
	 double *fopt, int *nup, int *ndown, int *nrej, int *lnobds, int *nnew)
{
  /* System generated locals */
  
  double d_1;
  
  /* Local variables */
  
  static int totmov;
  
  /* Subroutine */   
  
  extern int prtvec(double *, int *, char *);
  
  /* Parameter adjustments */
  
  --vm;
  --xopt;
  
  /* Function Body */
  
  totmov = *nup + *ndown + *nrej;
  
  printf("INTERMEDIATE RESULTS BEFORE NEXT TEMPERATURE REDUCTION \n");
  printf("CURRENT TEMPERATURE: %lf \n", *t);
  
  if(*max_) {
    printf("MAX FUNCTION VALUE SO FAR:   %lf \n", *fopt);
    printf("TOTAL MOVES:                 %d \n", totmov);
    printf("UPHILL:                      %d \n", *nup);
    printf("ACCEPTED DOWNHILL:           %d \n", *ndown);
    printf("REJECTED DOWNHILL:           %d \n", *nrej);
    printf("OUT OF BOUNDS TRIALS:        %d \n", *lnobds);
    printf("NEW MAXIMA THIS TEMPERATURE: %d \n", *nnew);
  } else {
    d_1 = -(*fopt);
    
    printf("MIN FUNCTION VALUE SO FAR:   %lf \n", d_1);
    printf("TOTAL MOVES:                 %d \n", totmov);
    printf("UPHILL:                      %d \n", *nup);
    printf("ACCEPTED DOWNHILL:           %d \n", *ndown);
    printf("REJECTED DOWNHILL:           %d \n", *nrej);
    printf("OUT OF BOUNDS TRIALS:        %d \n", *lnobds);
    printf("NEW MINIMA THIS TEMPERATURE: %d \n", *nnew);
  }
  
  prtvec(&xopt[1], n, "CURRENT OPTIMAL X");
  prtvec(&vm[1], n, "STEP LENGTH (VM)");
  printf(" *** \n");
  return(0);
} /* prt9 */

/* Subroutine */ 

int prt10(void)
{
  printf("SA ACHIEVED TERMINATION CRITERIA. IER = 0. \n");
  return(0);
} /* prt10 */

/* Subroutine */ 

/*  This subroutine prints the double precision vector named VECTOR.    */
/*  Elements 1 thru NCOLS will be printed. NAME is a character variable */
/*  that describes VECTOR. Note that if NAME is given in the call to    */
/*  PRTVEC, it must be enclosed in quotes. If there are more than 10    */
/*  elements in VECTOR, 10 elements will be printed on each line.       */

int prtvec(double *vector, int *ncols, char *name_)
{
  /* System generated locals */
  
  int i_1, i_2;
  
  /* Local variables */
  
  static int i_, j, ll, lines;
  
  /* Parameter adjustments */
  
  --vector;
  
  /* Function Body */
  
  printf("%s \n", name_);
  
  if(*ncols > 10) {
    lines = (int) (*ncols / 10.f);
    i_1 = lines;
    
    for(i_ = 1; i_ <= i_1; ++i_) {
      ll = (i_ - 1) * 10;
      i_2 = ll + 10;
      
      for(j = ll + 1; j <= i_2; ++j) {
	printf("%lf \n", vector[j]);
      }
    }
    
    i_1 = *ncols;
    
    for(j = ll + 11; j <= i_1; ++j) {
      printf("%lf \n", vector[j]);
    }
  } else {
    i_1 = *ncols;
    
    for(j = 1; j <= i_1; ++j) {
      printf("%lf \n", vector[j]);
    }
  }
  
  return(0);
} /* prtvec */

